#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include <stdio.h>
#include <dtrace.h>

typedef PerlIO *OutputStream;

typedef struct dtrace_lib {
	dtrace_hdl_t *hdl;
} dtrace_lib_t;
typedef dtrace_lib_t *Sun__DTraceLib;

static SV *g_callback = (SV*)NULL;

static void
_callback(const char *line)
{
	dSP;

	PUSHMARK(SP);
	XPUSHs(sv_2mortal(newSVpv(line, 0)));
	PUTBACK;


	/* Call the Perl sub to process the callback */
	call_sv(g_callback, G_DISCARD);
}


static int
_dtrace_dcmdbuffered(const dtrace_bufdata_t *bufdata, void *arg)
{
	if(bufdata->dtbda_aggdata) {
		static char aggr_line[256] = {0};
		if (bufdata->dtbda_recdesc == NULL) {
			strcat(aggr_line, "\n");
			_callback(aggr_line);
			*aggr_line = 0;
		} else {
			if (*aggr_line == 0) {
				strcat(aggr_line, "aggr_");
				strcat(aggr_line, bufdata->dtbda_aggdata->dtada_desc->dtagd_name);
				strcat(aggr_line, ": ");
			}
			strcat(aggr_line, bufdata->dtbda_buffered);
		}
	} else
		_callback(bufdata->dtbda_buffered);
	return(DTRACE_HANDLE_OK);
}

static int
_chew(const dtrace_probedata_t *data, void *arg)
{

	return (DTRACE_CONSUME_THIS);
}

static int
_chewrec(const dtrace_probedata_t *data, const dtrace_recdesc_t *rec,
	void *arg)
{
	if (rec == NULL) {
		return (DTRACE_CONSUME_NEXT);
	}
	if (rec->dtrd_action == DTRACEACT_EXIT) {
		return (DTRACE_CONSUME_NEXT);
	}

	return (DTRACE_CONSUME_THIS);
}

MODULE = Sun::DTraceLib              PACKAGE = Sun::DTraceLib

VERSIONCHECK: DISABLE
PROTOTYPES: ENABLE
BOOT:
    {
    }

Sun::DTraceLib
new(package, fn)
	char *package
	SV *fn;
INIT:
	int err;
CODE:
	/* Remember the Perl sub */
	if (g_callback == (SV*)NULL)
		g_callback = newSVsv(fn);
	else
		SvSetSV(g_callback, fn);

	RETVAL = malloc(sizeof(dtrace_lib_t));

	RETVAL->hdl = dtrace_open(DTRACE_VERSION, 0, &err);
	if (RETVAL->hdl == NULL) {
		free(RETVAL);
		Perl_croak(aTHX_ "Cannot open dtrace library: %s\n",
			dtrace_errmsg(NULL, err));
	}

	if (dtrace_handle_buffered(RETVAL->hdl, _dtrace_dcmdbuffered, NULL) == -1) {
		dtrace_close(RETVAL->hdl);
		free(RETVAL);
		Perl_croak(aTHX_ "Couldn't add buffered handler");
	}

	/*
	* DTrace requires symbol resolution to be
	* switched on explicitly
	*/
#ifdef __APPLE__
	(void) dtrace_setopt(RETVAL->hdl, "stacksymbols", "enabled");
#endif

	/* always request flowindent information */
	(void) dtrace_setopt(RETVAL->hdl, "flowindent", 0);

	/* default buf/arg sizes */
	(void) dtrace_setopt(RETVAL->hdl, "bufsize", "1m");
	(void) dtrace_setopt(RETVAL->hdl, "aggsize", "1m");

	(void) dtrace_setopt(RETVAL->hdl, "stdc", "s");
OUTPUT:
	RETVAL

void
DESTROY(self)
	Sun::DTraceLib self;
CODE:
	dtrace_close(self->hdl);
	free(self);

dtrace_prog_t *
program_fcompile(self, filename)
	Sun::DTraceLib self;
	char *filename;
INIT:
	FILE *pfile;
CODE:
	pfile = fopen(filename, "r");
	if (pfile == NULL)
		Perl_croak(aTHX_ "Cannot open file %s", filename);

	RETVAL = dtrace_program_fcompile(self->hdl, pfile,
			0, 0, NULL);
	fclose(pfile);
	if (!RETVAL)
		Perl_croak(aTHX_ dtrace_errmsg(self->hdl, dtrace_errno(self->hdl)));
OUTPUT:
	RETVAL

dtrace_prog_t *
program_strcompile(self, script)
	Sun::DTraceLib self;
	char *script;
CODE:

	RETVAL = dtrace_program_strcompile(self->hdl, script,
			DTRACE_PROBESPEC_NAME, DTRACE_C_PSPEC, 1, NULL);
	if (!RETVAL)
		Perl_croak(aTHX_ dtrace_errmsg(self->hdl, dtrace_errno(self->hdl)));
OUTPUT:
	RETVAL

void
program_exec(self, prog)
	Sun::DTraceLib self;
	dtrace_prog_t *prog;
INIT:
	int ret;
	dtrace_proginfo_t info;
CODE:
	ret = dtrace_program_exec(self->hdl, prog, &info);
	if (ret != 0)
		Perl_croak(aTHX_ dtrace_errmsg(self->hdl, dtrace_errno(self->hdl)));

void
go(self)
	Sun::DTraceLib self;
	CODE:
	if (dtrace_go(self->hdl) < 0)
		Perl_croak(aTHX_ dtrace_errmsg(self->hdl, dtrace_errno(self->hdl)));

void
stop(self)
	Sun::DTraceLib self;
CODE:
	if (dtrace_stop(self->hdl) < 0)
		Perl_croak(aTHX_ dtrace_errmsg(self->hdl, dtrace_errno(self->hdl)));

int
status(self)
	Sun::DTraceLib self;
CODE:
	if ((RETVAL=dtrace_status(self->hdl)) < 0)
		Perl_croak(aTHX_ dtrace_errmsg(self->hdl, dtrace_errno(self->hdl)));
OUTPUT:
	RETVAL

int
work(self)
	Sun::DTraceLib self;
INIT:
	int ret;
CODE:
	ret=dtrace_work(self->hdl, NULL, _chew, _chewrec, NULL);
	RETVAL = (ret == DTRACE_WORKSTATUS_OKAY);
OUTPUT:
	RETVAL

void
sleep(self)
	Sun::DTraceLib self;
CODE:
	dtrace_sleep(self->hdl);

int
aggregate_print(self)
	Sun::DTraceLib self;
INIT:
	int ret;
CODE:
	ret=dtrace_aggregate_print(self->hdl, NULL, NULL);
	RETVAL = (ret == DTRACE_WORKSTATUS_OKAY);
OUTPUT:
	RETVAL

int
aggregate_snap(self)
	Sun::DTraceLib self;
CODE:
	RETVAL = dtrace_aggregate_snap(self->hdl);
OUTPUT:
	RETVAL

void
aggregate_clear(self)
	Sun::DTraceLib self;
CODE:
	dtrace_aggregate_clear(self->hdl);
